;;;; command.scm -- irc command handling

;;; Created:    <2002-07-12 23:34:44 foof>
;;; Time-stamp: <2002-07-28 21:41:28 foof>
;;; Author:     Alex Shinn <foof@synthcode.com>

(define-module irc.command
  (use irc.log)
  (use irc.reply)
  (use irc.user)
  (use irc.channel)
  (use gauche.net)
  (use gauche.regexp)
  (use gauche.charconv)
  (use srfi-1)
  (use srfi-2)
  (use srfi-11)
  (use srfi-13)
  (use srfi-14)
  (export irc-input irc-flush define-irc-command)
  )
(select-module irc.command)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; variables

(define irc-hostname "localhost")

(define irc-servername "localhost")

(define irc-server-info
  "Welcome to the Internet Relay Network
Your host is localhost, running version 0.0.(alpha)
This server was created Tue Jul  9 2002 at 09:20:20 JST
localhost 0.0.(alpha) dioswkg biklmnopstv")

(define irc-ops "NICKLEN=9 TOPICLEN=160 KICKLEN=160 CHANTYPES=+#&")

(define irc-motd
  "------------------------------------------------------------------------
 Welcome to the lain Wiki IRC.
------------------------------------------------------------------------")

;(define irc-motd (call-with-input-file "/home/foof/lang/gauche/irc/motd" port->string))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; general utilities

(define is-false-rx (string->regexp "^(0|no?|off|f(alse))?$"))
(define (is-false? arg) (rxmatch arg is-false-rx))
(define (is-true? arg) (not (is-false? arg)))

(define (irc-in-target? target user)
  (wlog "(irc-in-target? ~S ~S)" target user)
  (if (null? target)
    #f
    (let ((head (car target)))
      (or (if (irc-channel? head)
            (member head (map (lambda (c) (slot-ref (car c) 'name))
                              (slot-ref user 'channels)))
            (equal? head (slot-ref user 'name)))
          (irc-in-target? (cdr target) user)))))

(define (user-get-channel user ch)
  (wlog "(user-get-channel ~S ~S)" user ch)
  (let ((channels (slot-ref user 'channels)))
    (wlog "channels are: ~S" channels)
    (let ((res (find (lambda (x) (eq? (car x) ch)) channels)))
      (wlog "... done")
      res)))

(define (get-user ircd nick)
  (find (lambda (u) (string=? (slot-ref u 'name) nick))
        (slot-ref ircd 'clients)))

(define (irc-oper? user ch)
  (and-let* ((ch+perms (user-get-channel user ch)))
    (memq #\o (cdr ch+perms))))

(define (irc-nick user)
  (slot-ref user 'name))

(define (irc-nick&mode user ch)
  (and-let* ((name (slot-ref user 'name)))
    (if (and ch (irc-oper? user ch))
      (string-append "@" name)
      (string-append "+" name))))

(define (irc-user-id user)
  (and-let* ((name (slot-ref user 'name)))
    (format #f "~A!~A@~A" name name
            "localhost"
            ;;(slot-ref user 'hostname)
            )))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; parsing utilities

(define (irc-split str)
  (wlog "(irc-split ~S)" str)
  (let* ((colon (string-index str #\: 0)))
    (if colon
      (append (string-tokenize (substring str 0 colon))
              (list (substring str (+ 1 colon) (string-length str))))
      (string-tokenize str))))

(define (irc-split-target str)
  (string-split str #\,))

(define (irc-channel? str)
  (memq (string-ref str 0) '(#\# #\&)))

(define irc-comm-rx (string->regexp "^(:\\S+ )?(\\w+)(\\s+(.*))?"))

(define (handle-command ircd user command arg-str)
  (wlog "(handle-command ~S ~S ~S ~S)" ircd user command arg-str)
  (if (hash-table-exists? command-handlers command)
    ((hash-table-get command-handlers command)
     ircd user command (irc-split arg-str) arg-str)
    (begin
      (wlog "no such command: ~A\n" command)
      (irc-reply user ERR_UNKNOWNCOMMAND command))))

(define (irc-handle-request ircd user request)
  (wlog "(irc-handle-request ~S ~S ~S)\n" user request ircd)
  (rxmatch-if (rxmatch irc-comm-rx request)
     (#f prefix command #f rest)
   (handle-command ircd user (string->symbol (string-upcase command)) rest)
   (irc-reply user ERR_UNKNOWNCOMMAND request)))

(define command-line-rx (string->regexp "([^\r\n]*)\r?\n"))

(define (irc-input ircd user data)

  ;; log the input
  (ilog data)

  ;; append it to the input buffer
  (slot-set! user 'inbuffer (string-append (slot-ref user 'inbuffer) data))

  ;; Test whether the data in the buffer or the data we just read means
  ;; there is a complete request waiting to be fulfilled.  If there is,
  ;; handle it.
  (let* ((input (slot-ref user 'inbuffer))
         (init (if (string-incomplete? input)
                 (string-incomplete->complete input) input)))
    (let loop ((str init))
      (wlog "inbuffer: ~S\n" str)
      (let ((m (rxmatch command-line-rx str)))
        (if m
          (let ((line (rxmatch-substring m 1))
                (after (rxmatch-after m)))
            (slot-set! user 'inbuffer after)
            (wlog "got a command: ~S\n" line)
            (irc-handle-request ircd user line)
            (loop after))
          ;;(slot-set! user 'inbuffer str)
          ))))

  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; messaging utilities

(define (irc-raw-write user data)
  (olog data)
  (slot-set! user 'outbuffer (string-append (slot-ref user 'outbuffer) data)))

(define (irc-write user text)
  ;;(irc-raw-write user (ces-convert text "UTF-8" (slot-ref user 'charset)))
  (irc-raw-write user text)
  )

(define (irc-prompt user)
  (slot-set! user 'outbuffer (string-append (slot-ref user 'outbuffer)
                                            (slot-ref user 'prompt))))

(define (irc-flush user)
  (let ((port (slot-ref user 'output)))
    (if port
      (let ((text (slot-ref user 'outbuffer)))
        (unless (equal? text "")
                (display text port)
                (slot-set! user 'outbuffer "")))
      (wlog "user with no port"))))

(define irc-reply-arg-rx (string->regexp "<[^>]*>"))

(define (irc-reply user reply . args)
  (let ((code (car reply))
        (msg (cadr reply)))
    (irc-write user
     (format #f ":~A ~3'0D ~A ~A\r\n" irc-hostname code (irc-nick user)
             (regexp-replace-all irc-reply-arg-rx msg
                                 (lambda (m)
                                   (let ((a (car args)))
                                     (set! args (cdr args))
                                     a)))))))

(define (irc-notice user msg . args)
  (irc-write user (format #f ":~A NOTICE ~A :~A\r\n"
                          irc-hostname (irc-nick user)
                          (apply format #f msg args))))

(define (irc-welcome user ircd)
  (let ((i 1))
    (for-each
     (lambda (line)
       (irc-reply user (list i line))
       (set! i (+ i 1)))
     (string-split irc-server-info #\newline))
    (wlog "server info sent\n")
    (irc-reply user (list i "<ops> :are supported by this server") irc-ops)
    (irc-reply user RPL_LUSERCLIENT (length (slot-ref ircd 'clients)) 0 1)
    (irc-reply user RPL_LUSERME (length (slot-ref ircd 'clients)) 0)
    (irc-reply user RPL_MOTDSTART irc-hostname)
    (for-each (cut irc-reply user RPL_MOTD <>) (string-split irc-motd #\newline))
    (irc-reply user RPL_ENDOFMOTD)
    (irc-notice user "on 1 ca 2(4) ft 10(10) tr")))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; COMMANDS
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define command-handlers (make-hash-table))

(define (make-irc-proc args body)
  ;; split at :optional
  (let-values
   (((req opt) (break (cut eq? <> :optional) args)))
   (set! opt (if (null? opt) opt (cdr opt))) ;; drop the :optional
   (let* ((min-args (length req))
          (max-args (+ min-args (length opt))))
     ;;(wlog "(make-irc-proc ~S) => ~S ~S: min ~S, max ~S" args req opt min-args max-args)
     ;; real params for all irc commands
     `(lambda (ircd user command params request)
        ;; validate min arg count
        (if (< (length params) ,min-args)
          (irc-reply user ERR_NEEDMOREPARAMS command)
          ;; bind required args, if any
          (,@(if (null? req)
               `(let ((o params)))
               `(let*-values
                 (((r o) (split-at params ,min-args))
                  ((,@req) (apply values r)))))
           ;; validate max arg count
           (if (> (length params) ,max-args)
             (irc-notice user "too many arguments for ~S (~A max ~A)"
                         command (length params) ,max-args)
             ;; do we need to worry about additional bindings?
             ,(if (null? opt)
                `(begin ,@body) ;; no optional bindings
                `(let-values    ;; let-values does our work
                  (((,@opt)
                    (apply values (append o (make-list (- ,(length opt)
                                                          (length o)))))))
                  ,@body)))))))))

;; Wrap up the scheme procedure, command mapping, paramter checking, and
;; documentation for IRC commands in a single macro.  Confusing use of
;; multiple-values to achieve limited let-optional behaviour (no need
;; for :rest).
(define-macro (define-irc-command name commands args doc . body)
  (let ((procname (gensym))
        (proc (make-irc-proc args body)))
    ;;(format #t "~A is: ~S\n" name proc)
    `(define ,name
       (let ((,procname ,proc))
         (for-each
          (cut hash-table-put! command-handlers <> ,procname)
          ,(list 'quote `(,@commands)))
         ,procname))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; 4.1 Connection Registration
     
;; The commands described here are used to register a connection with an
;; IRC server as either a user or a server as well as correctly disconnect.

(define-irc-command handle-pass (PASS) (password)
  "The PASS command is used to set a 'connection password'. The
password can and must be set before any attempt to register the
connection is made. Currently this requires that clients send a PASS
command before sending the NICK/USER combination and servers *must* send
a PASS command before any SERVER command. The password supplied must
match the one contained in the C/N lines (for servers) or I lines (for
clients). It is possible to send multiple PASS commands before
registering but only the last one sent is used for verification and it
may not be changed once registered."
  (slot-set! user 'pass password))

(define-irc-command handle-nick (NICK) (nick :optional hopcount)
  "NICK message is used to give user a nickname or change the previous
one. The <hopcount> parameter is only used by servers to indicate
how far away a nick is from its home server. A local connection has
a hopcount of 0. If supplied by a client, it must be ignored."
  (if (any (lambda (u) (string-ci=? (or (slot-ref u 'name) "") nick))
           (slot-ref ircd 'clients))
    (irc-reply user ERR_NICKNAMEINUSE nick)
    (begin
      (slot-set! user 'name nick)
      (unless (slot-ref user 'welcomed)
              (slot-set! user 'welcomed #t)
              (irc-welcome user ircd)))))

(define-irc-command handle-user (USER) (username hostname servername realname)
  "The USER message is used at the beginning of connection to specify
the username, hostname, servername and realname of s new user. It is
also used in communication between servers to indicate new user
arriving on IRC, since only after both USER and NICK have been
received from a client does a user become registered."
  (slot-set! user 'user (list username hostname servername realname)))

(define-irc-command handle-server (SERVER) (servername hopcount info)
  "The server message is used to tell a server that the other end of a
new connection is a server. This message is also used to pass server
data over whole net. When a new server is connected to net, information
about it be broadcast to the whole network. <hopcount> is used to give
all servers some internal information on how far away all servers
are. With a full server list, it would be possible to construct a map of
the entire server tree, but hostmasks prevent this from being done."
  (irc-reply user ERR_UNKNOWNCOMMAND command))

(define-irc-command handle-oper (OPER) (user password)
  "OPER message is used by a normal user to obtain operator privileges.
The combination of <user> and <password> are required to gain
Operator privileges."
  (if (and (slot-ref user 'user)
           (slot-ref user 'pass))
    (slot-set! user 'oper #t)))

(define-irc-command handle-quit (QUIT) (:optional message)
  "A client session is ended with a quit message. The server must close
the connection to a client which sends a QUIT message. If a \"Quit
Message\" is given, this will be sent instead of the default message,
the nickname."
  ;;(socket-close (slot-ref user 'socket))
  (let* ((clients (slot-ref ircd 'clients))
         (remainder (delete user clients eq?)))
    (slot-set! ircd 'clients remainder)))

(define-irc-command handle-squit (SQUIT) (server :optional comment)
  "The SQUIT message is needed to tell about quitting or dead servers.
If a server wishes to break the connection to another server it must
send a SQUIT message to the other server, using the the name of the
other server as the server parameter, which then closes its
connection to the quitting server."
  (irc-reply user ERR_UNKNOWNCOMMAND command))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; 4.2 Channel operations

;; This group of messages is concerned with manipulating channels, their
;; properties (channel modes), and their contents (typically clients).
;; In implementing these, a number of race conditions are inevitable
;; when clients at opposing ends of a network send commands which will
;; ultimately clash. It is also required that servers keep a nickname
;; history to ensure that wherever a <nick> parameter is given, the
;; server check its history in case it has recently been changed.

(define (get-channel ircd channel)
  (wlog "(get-channel ~S ~S)" ircd channel)
  (and-let* ((chl (find (lambda (x) (string=? (slot-ref x 'name) channel))
                        (slot-ref ircd 'channels))))
    (wlog "found: ~S" chl)
    chl
    ))

(define (get-or-create-channel ircd channel)
  (let* ((channels (slot-ref ircd 'channels))
         (chl (find (lambda (c) (equal? channel (slot-ref c 'name)))
                    channels)))
    (if chl
      ;;(car chl)
      chl
      (let ((ch (make <irc-channel> :name channel)))
        (slot-set! ircd 'channels (cons ch channels))
        ch))))

(define (user-is-banned? ch user) #f)
(define (user-is-not-invited? ch user) #f)
(define (bad-channel-key? ch user) #f)
(define (channel-is-full? ch) #f)
(define max-channels 10)

(define-irc-command handle-join (JOIN) (channel/s :optional key/s)
  "The JOIN command is used by client to start listening a specific
channel. Whether or not a client is allowed to join a channel is checked
only by the server the client is connected to; all other servers
automatically add the user to the channel when it is received from other
servers."
  (wlog "handle-join")
  (let ((channels (irc-split-target channel/s))
        (keys (if key/s (irc-split-target key/s) '())))
    (for-each
     (lambda (channel)
       (let ((ch (get-or-create-channel ircd channel))
             (key (if (null? keys) #f (car keys))))
         (if (not (null? keys)) (set! keys (cdr keys)))
         (wlog "about to check perms: user => ~S, channel: ~S" user channel)
         (cond ((user-is-banned? ch user)
                (irc-reply user ERR_BANNEDFROMCHAN channel))
               ((user-is-not-invited? ch user)
                (irc-reply user ERR_INVITEONLYCHAN channel))
               ((bad-channel-key? ch key)
                (irc-reply user ERR_BADCHANNELKEY channel))
               ((channel-is-full? ch)
                (irc-reply user ERR_CHANNELISFULL channel))
               ((>= (length (slot-ref user 'channels)) max-channels)
                (irc-reply user ERR_TOOMANYCHANNELS channel))
               (else
                ;; ok to join channel
                (wlog "perms ok for channel ~S" ch)
                (let ((chs (slot-ref user 'channels))
                      (modes '()))
                  (unless (any (lambda (x) (eq? (car x) ch)) chs)
                    (wlog "in unless: users => ~S" (slot-ref ch 'users))
                    ;; if new (empty) channel, we're the op
                    (if (null? (slot-ref ch 'users))
                      (set! modes (cons #\o modes))))
                  ;; add to user's channel list
                  (slot-set! user 'channels (cons (cons ch modes) chs))
                  (wlog "past slot-set!")
                  ;; add to channel's user list
                  (let ((users (cons (cons user modes)
                                     (slot-ref ch 'users)))
                        (uid (irc-user-id user)))
                    (slot-set! ch 'users users)
                    (wlog "past 2nd slot-set!")
                    ;; notify all channel users
                    (for-each
                     (lambda (u)
                       (wlog "notifying user ~S (~S) of join" (slot-ref (car u) 'name) u)
                       (irc-write (car u) (format #f ":~A JOIN :~A\r\n"
                                                  uid channel)))
                     users))
                  ;; already joined or not, send back the names
                  (wlog "about to call names")
                  (handle-names ircd user 'names (list channel) "")
                  )))))
     channels)))

(define-irc-command handle-part (PART) (channel/s)
  "The PART message causes the client sending the message to be removed
from the list of active users for all given channels listed in the
parameter string."
  (let ((channels (irc-split-target channel/s)))
    (for-each
     (lambda (channel)
       (slot-set! user 'channels (delete (car params)
                                         (slot-ref user 'channels))))
     channels)))

(define (irc-send-mode user channel)
  (let* ((ch (user-get-channel user channel))
         (perm (if (pair? ch) (car ch) "+")))
    ;;(irc-reply RPL_UMODEIS (car ch))
    (irc-reply user RPL_CHANNELMODEIS channel perm)
    (irc-reply user '(329 "<channel> <number>") channel (sys-time))))

(define (irc-send-ban-list user channel)
  (irc-reply user RPL_ENDOFBANLIST channel))

(define-irc-command handle-mode (MODE) (channel :optional flags limit username ban-mask)
  "The MODE command is a dual-purpose command in IRC. It allows both
usernames and channels to have their mode changed. The rationale for
this choice is that one day nicknames will be obsolete and the
equivalent property will be the channel."
  (wlog "mode: ~A ~A" channel flags)
  (if (not flags)
    ;; nothing to change, just send the current mode
    (irc-send-mode user channel)
    ;; handle flags
    (let ((c (string-ref flags 0)))
      (wlog "handling specific mode")
      (cond ((eq? c #\+)
             ;; add settings (XXXX just echoing)
             (wlog "mode change ~A" flags)
             (let ((msg (format #f ":~A MODE ~A ~A\r\n"
                                (irc-nick user) ;;(irc-user-id user)
                                channel flags)))
               (wlog "message: ~A" msg)
               (irc-write user msg)
               (wlog "done")))
            ((eq? c #\-)
             ;; remove settings (XXXX just echoing)
             (irc-write user (format #f ":~A MODE ~A ~A\r\n"
                                     (irc-nick user) ;;(irc-user-id user)
                                     channel flags)))
            (else
             ;; query
             (if (equal? flags "b")
               (irc-send-ban-list user channel)
               (irc-send-mode user channel)))))))

(define (chanops-needed-to-set-topic? ch)
  (memq #\t (slot-ref ch 'modes)))

(define-irc-command handle-topic (TOPIC) (channel :optional topic)
  "The TOPIC message is used to change or view the topic of a channel.
The topic for channel <channel> is returned if there is no <topic>
given. If the <topic> parameter is present, the topic for that channel
will be changed, if the channel modes permit this action."
  ;;(wlog "handle-topic")
  (let* ((ch (get-channel ircd channel))
         (ch+perms (user-get-channel user ch)))
    ;;(wlog "got perms: ~S" ch+perms)
    (if (not ch+perms)
      (irc-reply user ERR_NOTONCHANNEL channel)
      (if (not topic)
        ;; no topic to set, reply current topic
        (let ((tp (slot-ref ch 'topic)))
          (wlog "no topic")
          (if tp
            (irc-reply user RPL_TOPIC channel topic)
            (irc-reply user RPL_NOTOPIC channel))
          (wlog "no topic done"))
        ;; set a topic
        (if (and (chanops-needed-to-set-topic? ch)
                 (not (irc-oper? user ch)))
          (irc-reply user ERR_CHANOPRIVSNEEDED channel)
          (let ((tp (slot-ref ch 'topic)))
            ;;(wlog "setting topic: ~S" tp)
            (if (not tp)
              (irc-write user (format #f ":~A TOPIC ~A :~A" (irc-user-id user)
                                      channel topic))
              (begin
                ;;(wlog "topic is: ~S" topic)
                (slot-set! ch 'topic topic)
                (irc-reply user RPL_TOPIC channel topic))))))))
  ;;(wlog "topic done")
  )

(define-irc-command handle-names (NAMES) (:optional channel)
  "By using the NAMES command, a user can list all nicknames that are
visible to them on any channel that they can see. Channel names
which they can see are those which aren't private (+p) or secret (+s)
or those which they are actually on. The <channel> parameter
specifies which channel(s) to return information about if valid.
There is no error reply for bad channel names."
  (let ((ch (get-channel ircd channel)))
    (wlog "channel: ~S" ch)
    (for-each
     (lambda (x)
       (and-let* ((n (irc-nick&mode x ch)))
         (wlog "user in channel: ~S" x)
         (irc-reply user RPL_NAMREPLY channel n)))
     (slot-ref ircd 'clients))
    (irc-reply user RPL_ENDOFNAMES channel)))

(define-irc-command handle-list (LIST) (channel/s :optional server)
  "The list message is used to list channels and their topics. If the
<channel> parameter is used, only the status of that channel is
displayed. Private channels are listed (without their topics) as channel
\"Prv\" unless the client generating the query is actually on that
channel. Likewise, secret channels are not listed at all unless the
client is a member of the channel in question."
  (irc-reply user ERR_UNKNOWNCOMMAND command))

(define-irc-command handle-invite (INVITE) (nickname channel)
  "The INVITE message is used to invite users to a channel. The
parameter <nickname> is the nickname of the person to be invited to
the target channel <channel>. There is no requirement that the
channel the target user is being invited to must exist or be a valid
channel. To invite a user to a channel which is invite only (MODE
+i), the client sending the invite must be recognised as being a
channel operator on the given channel."
  (irc-reply user ERR_UNKNOWNCOMMAND command))

(define-irc-command handle-kick (KICK) (channel user :optional comment)
  "The KICK command can be used to forcibly remove a user from a
channel. It 'kicks them out' of the channel (forced PART)."
  (irc-reply user ERR_UNKNOWNCOMMAND command))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; 4.3 Server queries and commands

;; The server query group of commands has been designed to return
;; information about any server which is connected to the network. All
;; servers connected must respond to these queries and respond
;; correctly. Any invalid response (or lack thereof) must be considered
;; a sign of a broken server and it must be disconnected/disabled as
;; soon as possible until the situation is remedied.

(define-irc-command handle-version (VERSION) (server)
  "The VERSION message is used to query the version of the server
program. An optional parameter <server> is used to query the version
of the server program which a client is not directly connected to."
  (irc-reply user ERR_UNKNOWNCOMMAND command))

(define-irc-command handle-stats (STATS) (query :optional server)
  "The stats message is used to query statistics of certain server. If
<server> parameter is omitted, only the end of stats reply is sent
back. The implementation of this command is highly dependent on the
server which replies, although the server must be able to supply
information as described by the queries below (or similar)."
  (irc-reply user ERR_UNKNOWNCOMMAND command))

;; The currently supported queries are:

;; c - returns a list of servers which the server may connect
;; to or allow connections from;
;; h - returns a list of servers which are either forced to be
;; treated as leaves or allowed to act as hubs;
;; i - returns a list of hosts which the server allows a client
;; to connect from;
;; k - returns a list of banned username/hostname combinations
;; for that server;
;; l - returns a list of the server's connections, showing how
;; long each connection has been established and the traffic
;; over that connection in bytes and messages for each
;; direction;
;; m - returns a list of commands supported by the server and
;; the usage count for each if the usage count is non zero;
;; o - returns a list of hosts from which normal clients may
;; become operators;
;; y - show Y (Class) lines from server's configuration file;
;; u - returns a string showing how long the server has been up.

(define-irc-command handle-links (LINKS) (server-mask :optional remote)
  "With LINKS, a user can list all servers which are known by the server
answering the query. The returned list of servers must match the
mask, or if no mask is given, the full list is returned."
  (irc-reply user ERR_UNKNOWNCOMMAND command))

(define-irc-command handle-time (TIME) (server)
  "The time message is used to query local time from the specified
server. If the server parameter is not given, the server handling the
command must reply to the query."
  (irc-reply user ERR_UNKNOWNCOMMAND command))

(define-irc-command handle-connect (CONNECT) (target-server :optional port remote-server)
  "The CONNECT command can be used to force a server to try to establish
a new connection to another server immediately. CONNECT is a
privileged command and is to be available only to IRC Operators. If
a remote server is given then the CONNECT attempt is made by that
server to <target server> and <port>."
  (irc-reply user ERR_UNKNOWNCOMMAND command))

(define-irc-command handle-trace (TRACE) (server)
  "TRACE command is used to find the route to specific server. Each
server that processes this message must tell the sender about it by
sending a reply indicating it is a pass-through link, forming a chain
of replies similar to that gained from using \"traceroute\". After
sending this reply back, it must then send the TRACE message to the
next server until given server is reached. If the <server> parameter
is omitted, it is recommended that TRACE command send a message to
the sender telling which servers the current server has direct
connection to."
  (irc-reply user ERR_UNKNOWNCOMMAND command))

(define-irc-command handle-admin (ADMIN) (server)
  "The admin message is used to find the name of the administrator of
the given server, or current server if <server> parameter is omitted.
Each server must have the ability to forward ADMIN messages to other
servers."
  (irc-reply user ERR_UNKNOWNCOMMAND command))

(define-irc-command handle-info (INFO) (server)
  "The INFO command is required to return information which describes
the server: its version, when it was compiled, the patchlevel, when
it was started, and any other miscellaneous information which may be
considered to be relevant."
  (irc-reply user ERR_UNKNOWNCOMMAND command))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; 4.4 Sending messages

;; The main purpose of the IRC protocol is to provide a base for clients
;; to communicate with each other. PRIVMSG and NOTICE are the only
;; messages available which actually perform delivery of a text message
;; from one client to another - the rest just make it possible and try
;; to ensure it happens in a reliable and structured manner.

(define-irc-command handle-privmsg (PRIVMSG NOTICE) (receiver text)
  "PRIVMSG is used to send private messages between users. <receiver>
is the nickname of the receiver of the message. <receiver> can also
be a list of names or channels separated with commas.

The NOTICE message is used similarly to PRIVMSG. The difference between
NOTICE and PRIVMSG is that automatic replies must never be sent in
response to a NOTICE message. This rule applies to servers too - they
must not send any error reply back to the client on receipt of a
notice. The object of this rule is to avoid loops between a client
automatically sending something in response to something it
received. This is typically used by automatons (clients with either an
AI or other interactive program controlling their actions) which are
always seen to be replying lest they end up in a loop with another
automaton."
  (let ((name (slot-ref user 'name)))
    (if (not name)
      (irc-reply user ERR_NOTREGISTERED)
      (let* ((targets (irc-split-target receiver))
             (msg (format #f ":~A PRIVMSG ~A :~A\r\n" (irc-user-id user)
                          receiver text)))
        (for-each
         (cut irc-write <> msg)
         (filter (lambda (u)
                   (and (not (eq? u user))
                        (irc-in-target? targets u)))
                 (slot-ref ircd 'clients)))))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; 4.5 User based queries

;; User queries are a group of commands which are primarily concerned
;; with finding details on a particular user or group users. When using
;; wildcards with any of these commands, if they match, they will only
;; return information on users who are 'visible' to you. The visibility
;; of a user is determined as a combination of the user's mode and the
;; common set of channels you are both on.

(define-irc-command handle-who (WHO) (channel :optional opers-only?)
  "The WHO message is used by a client to generate a query which returns
a list of information which 'matches' the <name> parameter given by the
client. In the absence of the <name> parameter, all visible users are
listed. The same result can be achieved by using a <name> of 0 or any
wildcard which will end up matching every entry possible."
  (for-each
   (lambda (x)
     ;;(and-let* ((n (irc-nick&mode x (get-channel ircd channel))))
     (and-let* ((n (irc-nick x)))
       (irc-reply user RPL_WHOREPLY channel (irc-nick user)
                  irc-hostname irc-servername n "@H" 0 "unknown")))
   (slot-ref ircd 'clients))
  (irc-reply user RPL_ENDOFWHO channel))

(define-irc-command handle-whois (WHOIS) (server :optional nick/s)
  "This message is used to query information about particular user. The
server will answer this message with several numeric messages indicating
different statuses of each user which matches the nickmask. If no
wildcard is present in the <nickmask>, any information about that nick
which you are allowed to see is presented. A comma (',') separated list
of nicknames may be given."
  (irc-reply user ERR_UNKNOWNCOMMAND command))

(define-irc-command handle-whowas (WHOWAS) (nick :optional count server)
  "Whowas asks for information about a nickname which no longer exists.
This may either be due to a nickname change or the user leaving IRC.
In response to this query, the server searches through its nickname
history, looking for any nicks which are lexically the same (no wild
card matching here). The history is searched backward, returning the
most recent entry first. If there are multiple entries, up to
<count> replies will be returned (or all of them if no <count>
parameter is given). If a non-positive number is passed as being
<count>, then a full search is done."
  (irc-reply user ERR_UNKNOWNCOMMAND command))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; 4.6 Miscellaneous messages

;; Messages in this category do not fit into any of the above categories
;; but are nonetheless still a part of and required by the protocol.

(define-irc-command handle-kill (KILL) (nickname :optional comment)
  "The KILL message is used to cause a client-server connection to be
closed by the server which has the actual connection. KILL is used
by servers when they encounter a duplicate entry in the list of valid
nicknames and is used to remove both entries. It is also available
to operators."
  (irc-reply user ERR_UNKNOWNCOMMAND command))

(define-irc-command handle-ping (PING) (server :optional message)
  "The PING message is used to test the presence of an active client at
the other end of the connection. A PING message is sent at regular
intervals if no other activity detected coming from a connection. If a
connection fails to respond to a PING command within a set amount of
time, that connection is closed.

PONG message is a reply to ping message. If parameter <daemon2> is given
this message must be forwarded to given daemon. The <daemon> parameter
is the name of the daemon who has responded to PING message and
generated this message."
  (let ((response (if (eq? command 'PING) "PONG" "PING"))
        (msg (if message (string-append " " message) "")))
    (irc-write user (format #f "~A~A\r\n" response msg))))

(define-irc-command handle-pong (PONG) (:optional server message)
  "PONG message is a reply to ping message. If parameter <daemon2> is given
this message must be forwarded to given daemon. The <daemon> parameter
is the name of the daemon who has responded to PING message and
generated this message."
  #f)

(define-irc-command handle-error (ERROR) (error-message)
  "The ERROR command is for use by servers when reporting a serious or
fatal error to its operators. It may also be sent from one server to
another but must not be accepted from any normal unknown clients."
  (irc-reply user ERR_UNKNOWNCOMMAND command))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; 5. OPTIONALS

;; This section describes OPTIONAL messages. They are not required in a
;; working server implementation of the protocol described herein. In
;; the absence of the option, an error reply message must be generated
;; or an unknown command error. If the message is destined for another
;; server to answer then it must be passed on (elementary parsing
;; required) The allocated numerics for this are listed with the
;; messages below.

(define-irc-command handle-away (AWAY) (message)
  "With the AWAY message, clients can set an automatic reply string for
any PRIVMSG commands directed at them (not to a channel they are on).
The automatic reply is sent by the server to client sending the PRIVMSG
command. The only replying server is the one to which the sending client
is connected to."
  (irc-reply user ERR_UNKNOWNCOMMAND command))

(define-irc-command handle-rehash (REHASH) ()
  "The rehash message can be used by the operator to force the server to
re-read and process its configuration file."
  (irc-reply user ERR_UNKNOWNCOMMAND command))

(define-irc-command handle-restart (RESTART) ()
  "The restart message can only be used by an operator to force a server
restart itself. This message is optional since it may be viewed as a
risk to allow arbitrary people to connect to a server as an operator
and execute this command, causing (at least) a disruption to service."
  (irc-reply user ERR_UNKNOWNCOMMAND command))

(define-irc-command handle-summon (SUMMON) (user :optional server)
  "The SUMMON command can be used to give users who are on a host
running an IRC server a message asking them to please join IRC. This
message is only sent if the target server (a) has SUMMON enabled, (b)
the user is logged in and (c) the server process can write to the
user's tty (or similar)."
  (irc-reply user ERR_UNKNOWNCOMMAND command))

(define-irc-command handle-users (USERS) (:optional server)
  "The USERS command returns a list of users logged into the server in a
similar format to who(1), rusers(1) and finger(1). Some people may
disable this command on their server for security related reasons. If
disabled, the correct numeric must be returned to indicate this."
  (irc-reply user ERR_UNKNOWNCOMMAND command))

(define-irc-command handle-wallops (WALLOPS) (message)
  "Sends a message to all operators currently online. After
implementing WALLOPS as a user command it was found that it was
often and commonly abused as a means of sending a message to a lot
of people (much similar to WALL). Due to this it is recommended
that the current implementation of WALLOPS be used as an
example by allowing and recognising only servers as the senders of
WALLOPS."
  (irc-reply user ERR_UNKNOWNCOMMAND command))

(define-irc-command handle-userhost (USERHOST) (nick1 :optional nick2 nick3 nick4 nick5)
  "The USERHOST command takes a list of up to 5 nicknames, each
separated by a space character and returns a list of information
about each nickname that it found. The returned list has each reply
separated by a space."
  (for-each
   (lambda (nick)
     (if nick
       (and-let* ((u (get-user ircd nick))
                  (userinfo (slot-ref u 'user)))
         (wlog "handle-userhost: found ~S ~S" u userinfo)
         (irc-reply user RPL_USERHOST nick (car userinfo)
                    ;;(cadddr userinfo)
                    "localhost")
         (wlog "handle-userhost: ok"))))
   (list nick1 nick2 nick3 nick4 nick5)))

(define-irc-command handle-ison (ISON) (nick1 :optional nick2 nick3 nick4 nick5)
  "The ISON command was implemented to provide a quick and efficient
means to get a response about whether a given nickname was currently
on IRC. ISON only takes one (1) parameter: a space-separated list of
nicks. For each nickname in the list that is present, the server
adds that to its reply string. Thus the reply string may return
empty (none of the given nicks are present), an exact copy of the
parameter string (all of them present) or as any other subset of the
set of nicks given in the parameter. The only limit on the number
of nicks that may be checked is that the combined length must not be
too large as to cause the server to chop it off so it fits in 512
characters."
  (irc-reply user ERR_UNKNOWNCOMMAND command))

(provide "irc/command")

