;;;; ircd.scm -- irc daemon

;;; Created:    <2002-07-07 13:22:09 foof>
;;; Time-stamp: <2002-07-13 18:32:28 foof>
;;; Author:     Alex Shinn <foof@synthcode.com>

(define-module irc.ircd
  (use gauche.net)
  (use gauche.logger)
  (use gauche.selector)
  (use irc.log)
  (use irc.user)
  (use irc.server)
  (use irc.command)
  (use srfi-13)
  (export <irc-server> make-ircd ircd-loop ircd-close)
  )
(select-module irc.ircd)

(define (sockaddr->string addr)
  (let* ((a (with-output-to-string (lambda () (write addr))))
         (q (or (string-index a #\") -1))
         (c (or (string-index a #\:) (string-length a))))
    (substring a (+ q 1) c)))

(define (sockaddr-name addr)
  (slot-ref (sys-gethostbyaddr (sockaddr->string addr) AF_INET) 'name))

(define (make-ircd . opt-port)

  (define selector (make <selector>))
  (define socket (make-server-socket
                  (if (null? opt-port) default-port (car opt-port))
                  :reuse-addr #t))
  (define ircd (make <irc-server> :selector selector :socket socket))
  ;;:hostname (sockaddr-name socket)

  (define (acceptor sock flag)
    (let* ((client (socket-accept socket))
           (output (socket-output-port client))
           (user (make <irc-user> :socket client :output output
                       :host (sockaddr-name client))))
      (slot-set! ircd 'clients (cons user (slot-ref ircd 'clients)))
      (selector-add! selector
                     (socket-input-port client :buffered? #f)
                     (lambda (input flag)
                       (handler user input output))
                     '(r))))

  (define (handler user input output)
    (with-error-handler
     (lambda (err)
       (wlog "error handling request: ~S" err))
     (lambda ()
       (let ((str (read-block 4096 input)))
         (cond ((eof-object? str)
                (selector-delete! selector input #f #f)
                (socket-close (slot-ref user 'socket)))
               (else
                (irc-input ircd user str)))
         (for-each irc-flush (slot-ref ircd 'clients))))))

  (selector-add! selector
                 (socket-fd socket)
                 acceptor
                 '(r))

  ircd )

(define (ircd-loop ircd)
  (do ()
      ((eq? 'closed (socket-status (slot-ref ircd 'socket))) #f)
    (selector-select (slot-ref ircd 'selector))))

(define (ircd-close ircd)
  (socket-shutdown (slot-ref ircd 'socket)))

(provide "irc/ircd")

